"""Module for classes that provide multiple credentials to authenticate on the API."""

import logging

from azure.identity import ClientSecretCredential, DefaultAzureCredential

# Get a logger for this module
logger = logging.getLogger(__name__)

# Scope for Azure DevOps
SCOPE = "499b84ac-1321-427f-aa17-267ca6975798"


class TokenCredential:
    """Base class for token based credentials."""

    def get_token(self):
        """Return a token."""
        raise NotImplementedError("Missing method implementation")


class AccessTokenCredential(TokenCredential):
    """Class that manage authentication through an access token."""

    def __init__(self, token: str):
        """Init a new access token credential."""
        logger.debug("Azure DevOps: using provided token to authenticate")
        self.token = token

    def get_token(self):
        """Return a token."""
        return self.token


class AzureCredential(TokenCredential):
    """Class that auto detect best authentication method on Azure."""

    def __init__(self):
        """Init a new default azure credential."""
        self.credentials = DefaultAzureCredential()

    def get_token(self):
        """Return a token."""
        return self.credentials.get_token(SCOPE).token


class ServicePrincipalCredential(TokenCredential):
    """Class that authenticate using a service principal."""

    def __init__(self, client_id: str, client_secret: str, tenant_id: str):
        """Init a new service principal credential."""
        self.credentials = ClientSecretCredential(client_id=client_id, client_secret=client_secret, tenant_id=tenant_id)

    def get_token(self):
        """Return a token."""
        return self.credentials.get_token(SCOPE).token


__all__ = [
    "AccessTokenCredential",
    "AzureCredential",
    "ServicePrincipalCredential",
    "TokenCredential",
]



"""Enumerations for Git Pull Requests types."""

from ...base.enums import StrIntEnum


class CommentTypeEnum(StrIntEnum):
    """Enum for available pull request thread comments."""

    unknown = 0
    text = 1
    codeChange = 2
    system = 3


class ThreadStatusEnum(StrIntEnum):
    """Enum for available pull request thread statuses."""

    unknown = 0
    active = 1
    fixed = 2
    wontFix = 3
    closed = 4
    byDesign = 5
    pending = 6



"""Module for Policy Teypes client API."""

import logging

from ...base.clients import Client
from .models import PolicyType, PolicyTypeCollection

# Get a logger for this module
logger = logging.getLogger(__name__)


class PolicyTypeClient(Client):
    """Represent a client to Policy Types API in Azure DevOps."""

    def list(self) -> PolicyTypeCollection:
        """List all policy types available in a project."""
        return PolicyTypeCollection.model_validate(self._client.get("").raise_for_status().json())

    def get(self, id: str) -> PolicyType:
        """Get a policy type by ID.."""
        return PolicyType.model_validate(self._client.get(id).raise_for_status().json())



"""Models for policy types schemas in Azure DevOps."""

from typing import Annotated

from pydantic import Field, HttpUrl

from ...base.models import BaseCollection, JSONModel


class PolicyType(JSONModel):
    """Represent a policy type configuration."""

    id: str
    url: HttpUrl
    display_name: Annotated[str, Field(alias="displayName")]
    description: str | None = None


class PolicyTypeCollection(BaseCollection[PolicyType]):
    """Represent a collection of policy types."""



"""Module for loading Azure DevOps API models."""

from .builds.models import (
    BuildDefinition,
    BuildDefinitionBase,
    BuildDefinitionCollection,
    BuildDefinitionCreate,
    BuildProcess,
    BuildRepository,
)
from .core.models import Project, Properties
from .git.models import (
    GitItem,
    GitItemCollection,
    GitItemDescriptor,
    GitItemsBatch,
    GitRef,
    GitRefCollection,
    GitRepository,
)
from .git.pullrequests.enums import CommentTypeEnum, ThreadStatusEnum
from .git.pullrequests.models import (
    PullRequestThread,
    PullRequestThreadCollection,
    PullRequestThreadComment,
    PullRequestThreadCommentCreate,
    PullRequestThreadCreate,
)
from .pipelines.models import (
    Pipeline,
    PipelineCollection,
    PipelineConfiguration,
    PipelineConfigurationRepository,
    PipelineCreate,
)
from .policy.configurations.models import (
    PolicyConfiguration,
    PolicyConfigurationCollection,
    PolicyConfigurationCreate,
    PolicyConfigurationUpdate,
    PolicyScope,
    PolicySettings,
    PolicyType,
)
from .servicehooks.subscriptions.models import HookSubscription, HookSubscriptionCollection, HookSubscriptionCreate

__all__ = [
    "BuildDefinition",
    "BuildDefinitionBase",
    "BuildDefinitionCollection",
    "BuildDefinitionCreate",
    "BuildProcess",
    "BuildRepository",
    "CommentTypeEnum",
    "GitItem",
    "GitItemCollection",
    "GitItemDescriptor",
    "GitItemsBatch",
    "GitRef",
    "GitRefCollection",
    "GitRepository",
    "HookSubscription",
    "HookSubscriptionCollection",
    "HookSubscriptionCreate",
    "Pipeline",
    "PipelineCollection",
    "PipelineConfiguration",
    "PipelineConfigurationRepository",
    "PipelineCreate",
    "PolicyConfiguration",
    "PolicyConfigurationCollection",
    "PolicyConfigurationCreate",
    "PolicyConfigurationUpdate",
    "PolicyScope",
    "PolicySettings",
    "PolicyType",
    "Project",
    "Properties",
    "PullRequestThread",
    "PullRequestThreadCollection",
    "PullRequestThreadComment",
    "PullRequestThreadCommentCreate",
    "PullRequestThreadCreate",
    "ThreadStatusEnum",
]



"""Azure DevOps Base Client."""

import atexit
import logging

import httpx

from .builds.clients import BuildClient
from .core.clients import ProjectClient
from .credentials import AzureCredential, TokenCredential
from .git.clients import GitPolicyConfigurationClient, GitRepositoryClient
from .git.pullrequests.clients import PullRequestClient
from .pipelines.clients import PipelineClient
from .policy.configurations.clients import PolicyConfigurationClient
from .policy.types.clients import PolicyTypeClient
from .servicehooks.subscriptions.clients import HookSubscriptionClient

# Get a logger for this module
logger = logging.getLogger(__name__)


class AzureDevOps:
    """Represent an Azure DevOps connection."""

    API_VERSION = "7.1"

    def __init__(self, org_url: str):
        """Create a new Azure DevOps connection."""
        if not org_url.startswith("https://dev.azure.com/"):
            raise ValueError("You did not provide an URL to an Azure DevOps organization !")

        atexit.register(self._terminate)

        self.org_url = org_url
        self._token: str | None = None
        self._clients_cache: dict[int, httpx.Client] = {}

    def authenticate(self, credentials: TokenCredential | None = None):
        """
        Authenticate on Azure DevOps.

        If `credentials` is not provided, will use default [Azure authentication](https://learn.microsoft.com/en-us/dotnet/azure/sdk/authentication/credential-chains?tabs=dac#defaultazurecredential-overview).
        """
        if not credentials:
            credentials = AzureCredential()
        self._token = credentials.get_token()

    def _build_client(self, endpoint: str) -> httpx.Client:
        """Return an HTTP client for interacting with an Azure DevOps API endpoint."""
        default_params = {"api-version": AzureDevOps.API_VERSION}
        client = httpx.Client(base_url=endpoint, timeout=5.0, params=default_params)
        if self._token:
            client.headers.update({"Authorization": f"Bearer {self._token}"})
        else:
            raise ValueError(
                "You are not connected to Azure DevOps ! Call one of *_authenticate() methods first to get a token."
            )
        fingerprint = hash((client.base_url, client.params))

        # Get a client from global cache or store a new one
        if fingerprint in self._clients_cache:
            existing_client = self._clients_cache[fingerprint]
            logger.debug("Found an existing client for %s (%d)", existing_client.base_url, fingerprint)
            return existing_client
        else:
            logger.debug("Create a new client in cache %s (%d)", client.base_url, fingerprint)
            self._clients_cache.update({fingerprint: client})
            return client

    def _terminate(self):
        """Terminate all client connections at exit."""
        logger.debug("Closing all Azure DevOps API clients...")
        for fingerprint, client in self._clients_cache.items():
            logger.debug("Terminating client for %s (%d).", client.base_url, fingerprint)
            client.close()

    def projects_client(self) -> ProjectClient:
        """Return an HTTP client for interacting with Projects endpoint."""
        _endpoint = [self.org_url, "_apis", "projects"]
        return ProjectClient(self._build_client(endpoint="/".join(_endpoint)))

    def pull_request_client(self, project: str, repository: str) -> PullRequestClient:
        """Return an HTTP client for interacting with Pull Request endpoint."""
        _endpoint = [self.org_url, project, "_apis", "git", "repositories", repository, "pullrequests"]
        return PullRequestClient(self._build_client(endpoint="/".join(_endpoint)))

    def git_repository_client(self, project: str | None = None) -> GitRepositoryClient:
        """Return an HTTP client for interacting with a Git repository endpoint."""
        _endpoint = [self.org_url, "_apis", "git", "repositories"]
        if project:
            _endpoint.insert(1, project)
        return GitRepositoryClient(self._build_client(endpoint="/".join(_endpoint)))

    def git_repository_policy_configuration_client(self, project: str) -> GitPolicyConfigurationClient:
        """Return an HTTP client for interacting with a Git Repository Policy Configuration endpoint."""
        _endpoint = [self.org_url, project, "_apis", "git", "policy", "configurations"]
        return GitPolicyConfigurationClient(self._build_client(endpoint="/".join(_endpoint)))

    def policy_configuration_client(self, project: str) -> PolicyConfigurationClient:
        """Return an HTTP client for interacting with Policy Configuration endpoint."""
        _endpoint = [self.org_url, project, "_apis", "policy", "configurations"]
        return PolicyConfigurationClient(self._build_client(endpoint="/".join(_endpoint)))

    def policy_types_client(self, project: str) -> PolicyTypeClient:
        """Return an HTTP client for interacting with Policy Types endpoint."""
        _endpoint = [self.org_url, project, "_apis", "policy", "types"]
        return PolicyTypeClient(self._build_client(endpoint="/".join(_endpoint)))

    def pipeline_client(self, project: str) -> PipelineClient:
        """Return an HTTP client for interacting with Pipelines endpoint."""
        _endpoint = [self.org_url, project, "_apis", "pipelines"]
        return PipelineClient(self._build_client(endpoint="/".join(_endpoint)))

    def builds_client(self, project: str) -> BuildClient:
        """Return an HTTP client for interacting with Builds endpoint."""
        _endpoint = [self.org_url, project, "_apis", "build"]
        return BuildClient(self._build_client(endpoint="/".join(_endpoint)))

    def hook_subscriptions_client(self) -> HookSubscriptionClient:
        """Return an HTTP client for interacting with Service Hooks' subscriptions endpoint."""
        _endpoint = [self.org_url, "_apis", "hooks", "subscriptions"]
        return HookSubscriptionClient(self._build_client(endpoint="/".join(_endpoint)))



"""Models for policy configurations schemas in Azure DevOps."""

from typing import Annotated, Self

from pydantic import Field

from ...base.models import BaseCollection, JSONModel
from ..types.models import PolicyType


class PolicyScope(JSONModel):
    """Represent the scope of a policy."""

    repository_id: Annotated[str | None, Field(alias="repositoryId", default=None)] = None
    ref_name: Annotated[str, Field(alias="refName")]
    match_kind: Annotated[str, Field(alias="matchKind")]


class PolicySettings(JSONModel):
    """Represent settings for a policy configuration."""

    display_name: Annotated[str | None, Field(alias="displayName")]
    build_definition_id: Annotated[int, Field(alias="buildDefinitionId")]
    never_expire: Annotated[bool, Field(alias="queueOnSourceUpdateOnly", default=False)] = False
    manual_trigger: Annotated[bool, Field(alias="manualQueueOnly", default=False)] = False
    valid_duration: Annotated[int, Field(alias="validDuration", default=0)] = 0
    scope: list[PolicyScope]


class PolicyConfigurationBase(JSONModel):
    """Represent a request to create a policy configuration."""

    enabled: Annotated[bool, Field(alias="isEnabled")]
    required: Annotated[bool, Field(alias="isBlocking")]
    enterprise_managed: Annotated[bool, Field(alias="isEnterpriseManaged", default=False)] = False
    settings: PolicySettings
    type: PolicyType


class PolicyConfigurationCreate(PolicyConfigurationBase):
    """Represent a request to create a policy configuration."""


class PolicyConfigurationUpdate(PolicyConfigurationBase):
    """Represent a request to update a policy configuration."""


class PolicyConfiguration(PolicyConfigurationBase):
    """Represent a policy configuration."""

    id: int
    deleted: Annotated[bool, Field(alias="isDeleted")]


class PolicyConfigurationCollection(BaseCollection[PolicyConfiguration]):
    """Represent a collection of policy configuration."""

    def _is_build_type(self, policy: PolicyConfiguration) -> bool:
        return not policy.deleted and policy.type.id == "0609b952-1397-4640-95ec-e00a01b2c241"

    def get_build_policies(self) -> Self:
        """Return list of build policies from a collection of policy configuration."""

        def _filter(policy: PolicyConfiguration) -> bool:
            return self._is_build_type(policy)

        filtered_results = list(filter(_filter, self))
        return type(self)(count=len(filtered_results), value=filtered_results)

    def match_build_definition(self, id: int) -> Self:
        """Return list of build policies that match build definition by ID."""

        def _filter(policy: PolicyConfiguration) -> bool:
            if self._is_build_type(policy):
                return policy.settings.build_definition_id == id
            else:
                return False

        filtered_results = list(filter(_filter, self))
        return type(self)(count=len(filtered_results), value=filtered_results)



"""Models for the Pipelines Azure DevOps API endpoint."""

from pathlib import PurePosixPath, PureWindowsPath
from typing import Annotated, Literal, Self

from pydantic import BeforeValidator, Field, HttpUrl

from ..base import validators
from ..base.models import BaseCollection, JSONModel
from .enums import ConfigurationType


class Pipeline(JSONModel):
    """Represents a pipeline resource."""

    id: int
    revision: int
    name: str
    folder: Annotated[PureWindowsPath, BeforeValidator(validators.windows_path)]
    url: HttpUrl


class PipelineConfigurationRepository(JSONModel):
    """Represents a pipeline configuration repository."""

    id: str
    type: Literal["azureReposGit"] = "azureReposGit"


class PipelineConfiguration(JSONModel):
    """Represents configuration of a pipeline resource."""

    path: Annotated[PurePosixPath, BeforeValidator(validators.posix_path)]
    type: Annotated[
        ConfigurationType, Field(default=ConfigurationType.yaml), BeforeValidator(ConfigurationType.validate)
    ] = ConfigurationType.yaml
    repository: PipelineConfigurationRepository


class PipelineDefinition(Pipeline):
    """Represents a pipeline definiton."""

    configuration: PipelineConfiguration


class PipelineCreate(JSONModel):
    """Represents a request to create a pipeline resource."""

    name: str
    folder: Annotated[PureWindowsPath, BeforeValidator(validators.windows_path)]
    configuration: PipelineConfiguration


class PipelineCollection(BaseCollection[Pipeline]):
    """Represents a collection of pipeline resources."""

    def startswith(self, pattern: str) -> Self:
        """Filter pipelines that have a name starting with `pattern`."""

        def _filter(p: Pipeline) -> bool:
            return p.name.startswith(pattern)

        filtered_results = list(filter(_filter, self))
        return type(self)(count=len(filtered_results), value=filtered_results)

    def from_folder(self, pattern: str) -> Self:
        """Filter pipelines that are contained in a folder."""

        def _filter(p: Pipeline) -> bool:
            return p.folder.match(pattern)

        filtered_results = list(filter(_filter, self))
        return type(self)(count=len(filtered_results), value=filtered_results)



"""Test Service Hooks Models."""

import pydantic
import pytest

from ez_ados.models import HookSubscription, HookSubscriptionCollection, HookSubscriptionCreate


def test_hook_subscription_collection():
    """Test a collection of Service Hooks' Subscription resource."""
    spec = {
        "id": "123",
        "status": "enabled",
        "publisher_id": "tfs",
        "event_type": "git.push",
        "resource_version": "1.0-preview.1",
        "event_description": "Resource was updated",
        "consumer_id": "azureStorageQueue",
        "consumer_action_id": "enqueue",
        "action_description": "Consumer performed action",
        "publisher_inputs": {"branch": "test", "projectId": "1234"},
        "consumer_inputs": {"accountName": "stafoobar001"},
    }
    resources = [
        HookSubscription.model_validate(spec),
        HookSubscription.model_validate(spec | {"event_type": "git.repo.deleted", "status": "disabledBySystem"}),
    ]
    assert resources[0].status.value == 0  # noqa: PLR2004
    assert resources[1].status.value == 30  # noqa: PLR2004

    try:
        results = HookSubscriptionCollection.model_validate({"count": len(resources), "value": resources})
    except pydantic.ValidationError as exc:
        pytest.fail(reason=str(exc))

    with pytest.raises(pydantic.ValidationError):
        HookSubscription(id=1)

    filtered = results.for_git_push_event()
    assert filtered.count == 1  # noqa: PLR2004


def test_hook_subscription_create():
    """Test creation of a Service Hooks' Subscription resource."""
    spec = {
        "publisher_id": "tfs",
        "event_type": "git.push",
        "resource_version": "1.0-preview.1",
        "consumer_id": "azureStorageQueue",
        "consumer_action_id": "enqueue",
        "publisher_inputs": {"branch": "test", "projectId": "1234"},
        "consumer_inputs": {"accountName": "stafoobar001"},
    }
    resource = HookSubscriptionCreate.model_validate(spec)
    assert resource.publisher_id == "tfs"  # noqa: PLR2004
    assert resource.event_type == "git.push"  # noqa: PLR2004



"""Enumerations for Pipelines types."""

from ..base.enums import StrIntEnum


class ConfigurationType(StrIntEnum):
    """
    Enum for available configuration types.

    Reference: https://learn.microsoft.com/en-us/javascript/api/azure-devops-extension-api/configurationtype
    """

    unknown = 0
    yaml = 1
    designerHyphenJson = 2
    designerJson = 2
    justInTime = 3



"""Module for Service Hooks' Subscriptions models."""

from typing import Annotated, Any, Self

from pydantic import BeforeValidator, Field

from ...base.models import BaseCollection, JSONModel
from .enums import SubscriptionStatus


class HookSubscription(JSONModel):
    """Represents a hook subscription."""

    id: str
    status: Annotated[SubscriptionStatus, BeforeValidator(SubscriptionStatus.validate)]
    publisher_id: Annotated[str, Field(alias="publisherId")]
    event_type: Annotated[str, Field(alias="eventType")]
    resource_version: Annotated[str, Field(alias="resourceVersion")]
    event_description: Annotated[str, Field(alias="eventDescription")]
    consumer_id: Annotated[str, Field(alias="consumerId")]
    consumer_action_id: Annotated[str, Field(alias="consumerActionId")]
    action_description: Annotated[str, Field(alias="actionDescription")]
    publisher_inputs: Annotated[dict[str, Any], Field(alias="publisherInputs")]
    consumer_inputs: Annotated[dict[str, Any], Field(alias="consumerInputs")]


class HookSubscriptionCreate(JSONModel):
    """Creates a new hook subscription."""

    publisher_id: Annotated[str, Field(alias="publisherId")]
    event_type: Annotated[str, Field(alias="eventType")]
    resource_version: Annotated[str, Field(alias="resourceVersion")]
    consumer_id: Annotated[str, Field(alias="consumerId")]
    consumer_action_id: Annotated[str, Field(alias="consumerActionId")]
    publisher_inputs: Annotated[dict[str, Any], Field(alias="publisherInputs")]
    consumer_inputs: Annotated[dict[str, Any], Field(alias="consumerInputs")]


class HookSubscriptionCollection(BaseCollection[HookSubscription]):
    """Represents a collection of hook subscriptions."""

    def for_event(self, event_type: str) -> Self:
        """Filter hook subscription for a type of event."""

        def _filter(p: HookSubscription) -> bool:
            return p.event_type == event_type

        filtered_results = list(filter(_filter, self))
        return type(self)(count=len(filtered_results), value=filtered_results)

    def for_git_push_event(
        self, branch_name: str | None = None, project_id: str | None = None, repository_id: str | None = None
    ) -> Self:
        """Filter hook subscription for a git.push event."""
        _filtered_events = self.for_event("git.push")

        def _filter(p: HookSubscription) -> bool:
            if p.event_type == "git.push":
                if branch_name is not None and p.publisher_inputs["branch"] != branch_name:
                    return False
                if project_id is not None and p.publisher_inputs["projectId"] != project_id:
                    return False
                if repository_id is not None and p.publisher_inputs["repository"] != repository_id:
                    return False
                return True
            else:
                return False

        filtered_results = list(filter(_filter, _filtered_events))
        return type(self)(count=len(filtered_results), value=filtered_results)



"""Base classes for instantiating clients."""

import httpx


class Client:
    """Base class for a client."""

    def __init__(self, client: httpx.Client):
        """Instantiate a new client."""
        self._client = client
        self.base_url = self._client.base_url

    def close(self) -> None:
        """Close client connection."""
        self._client.close()



"""Enumerations for Git types."""

from ..base.enums import StrIntEnum


class RecursionType(StrIntEnum):
    """Enum for available recursion modes."""

    none = 0
    oneLevel = 1
    oneLevelPlusNestedEmptyFolders = 4
    full = 120


class GitObjectType(StrIntEnum):
    """Enum for available type of objects in git."""

    bad = 0
    commit = 1
    tree = 2
    blob = 3
    tag = 4
    ext2 = 5
    ofsDelta = 6
    refDelta = 7



"""Test Azure DevOps Models."""

import pydantic
import pytest

from ez_ados.models import (
    BuildDefinitionCreate,
    BuildProcess,
    BuildRepository,
    GitItem,
    GitItemCollection,
    GitItemDescriptor,
    GitItemsBatch,
    GitRef,
    GitRefCollection,
    GitRepository,
    Pipeline,
    PipelineCollection,
    PolicyConfiguration,
    PolicyConfigurationCollection,
    Project,
    PullRequestThread,
    PullRequestThreadCollection,
    PullRequestThreadCommentCreate,
    PullRequestThreadCreate,
)


def test_build_definition_create():
    """Test Build Definition resource instantiation."""
    resource = BuildDefinitionCreate(
        name="test",
        path="/ci",
        process=BuildProcess(yaml_filename="/my/folder/ci.yaml"),
        repository=BuildRepository(id="1234"),
    )
    assert resource.name == "test"

    with pytest.raises(pydantic.ValidationError):
        BuildDefinitionCreate(
            name="test",
            path="/ci",
            process="whatever",
            repository="1234",
        )


def test_git_item():
    """Test Git Item resource instantiation."""
    resources = [
        GitItem(objectId="1452645214266465", gitObjectType="blob", path="tests/folder"),
        GitItem(objectId="9809798755", gitObjectType="commit", path="tests/folder"),
    ]
    assert resources[0].git_object_type.value == 3  # noqa: PLR2004
    assert resources[1].git_object_type.value == 1  # noqa: PLR2004

    try:
        GitItemCollection.model_validate({"count": len(resources), "value": resources})
    except pydantic.ValidationError as exc:
        pytest.fail(reason=str(exc))

    with pytest.raises(pydantic.ValidationError):
        GitItem(objectId="1452645214266465", gitObjectType="foo")


def test_git_item_batch_body():
    """Test Git Item Batch Body resource instantiation."""
    resources = [
        GitItemDescriptor(path="/a/folder", version="main", recursionLevel="full"),
        GitItemDescriptor(path="/b/folder", version="main", recursionLevel="oneLevel"),
    ]
    assert resources[0].recursion_level.value == 120  # noqa: PLR2004
    assert resources[1].recursion_level.value == 1  # noqa: PLR2004

    try:
        GitItemsBatch.model_validate({"itemDescriptors": resources})
    except pydantic.ValidationError as exc:
        pytest.fail(reason=str(exc))

    with pytest.raises(pydantic.ValidationError):
        GitItemDescriptor(path="/a/folder", version="main", recursionLevel="foo")


def test_git_ref():
    """Test Git Ref resource instantiation."""
    resources = [
        GitRef(objectId="123456", name="refs/heads/main"),
        GitRef(objectId="7890", name="refs/heads/develop"),
    ]
    assert resources[0].name == "refs/heads/main"
    assert resources[1].name == "refs/heads/develop"

    try:
        GitRefCollection.model_validate({"count": len(resources), "value": resources})
    except pydantic.ValidationError as exc:
        pytest.fail(reason=str(exc))

    with pytest.raises(pydantic.ValidationError):
        GitRef(objectId="123456", name="refs/heads/main", isLocked="foo")


def test_project():
    """Test Project resource instantiation."""
    resource = Project(id="1234-456", name="my-project")
    assert resource.name == "my-project"

    with pytest.raises(pydantic.ValidationError):
        Project(id=1, name="my-project")


def test_git_repository():
    """Test Git Repository resource instantiation."""
    spec = {
        "id": "1234",
        "name": "my-repo",
        "defaultBranch": "refs/heads/main",
        "isDisabled": False,
        "isInMaintenance": False,
        "project": {"id": "1234", "name": "my-project"},
        "remoteUrl": "https://...",
        "webUrl": "https://...",
        "size": 0,
    }
    resource = GitRepository.model_validate(spec)
    assert resource.name == "my-repo"

    with pytest.raises(pydantic.ValidationError):
        GitRepository.model_validate(spec | {"id": 123})


def test_pipeline():
    """Test Pipeline resource instantiation."""
    resource = Pipeline(id=1, name="my-pipeline", folder="/my-folder", url="https://example.com", revision=1)
    assert resource.name == "my-pipeline"

    with pytest.raises(pydantic.ValidationError):
        Pipeline(id="foo", name="bad-pipeline", folder="a:bad:folder", url="https://example.com", revision=1)


def test_pipeline_collection():
    """Test collection of Pipelines instantiation."""
    try:
        pipelines = [
            Pipeline(id=1, name="my-ci", folder="/ci", url="https://example.com", revision=1),
            Pipeline(id=2, name="my-cd", folder="/cd", url="https://example.com", revision=1),
        ]
        collection = PipelineCollection.model_validate({"count": len(pipelines), "value": pipelines})
    except pydantic.ValidationError as exc:
        pytest.fail(reason=str(exc))

    assert collection.startswith("my-")[0].name == "my-ci"
    assert collection.from_folder("/cd")[0].name == "my-cd"


def test_policy_configuration():
    """Test Policy Configuration resource instantiation."""
    spec = {
        "id": 1,
        "isEnabled": True,
        "isBlocking": True,
        "isDeleted": False,
        "isEnterpriseManaged": False,
        "settings": {
            "displayName": "Build",
            "buildDefinitionId": 10,
            "scope": [{"repositoryId": "1234", "refName": "refs/heads/main", "matchKind": "commit"}],
        },
        "type": {"id": "1123-4556", "url": "https://...", "displayName": "Build", "description": "hey"},
    }
    resources = [
        PolicyConfiguration.model_validate(spec),
        PolicyConfiguration.model_validate(spec | {"id": 2}),
    ]
    assert resources[0].settings.display_name == "Build"

    try:
        PolicyConfigurationCollection.model_validate({"count": len(resources), "value": resources})
    except pydantic.ValidationError as exc:
        pytest.fail(reason=str(exc))

    with pytest.raises(pydantic.ValidationError):
        PolicyConfiguration.model_validate(spec | {"id": "foo"})


def test_git_pull_request_thread():
    """Test Git Pull Request Thread resource instantiation."""
    spec = {
        "id": 1,
        "publishedDate": "2025-10-03T12:46:41.42Z",
        "lastUpdatedDate": "2025-10-03T12:46:41.42Z",
        "isDeleted": False,
        "comments": [
            {
                "id": 1,
                "parentCommentId": 0,
                "content": "Policy status has been updated",
                "publishedDate": "2025-10-03T12:46:41.42Z",
                "lastUpdatedDate": "2025-10-03T12:46:41.42Z",
                "lastContentUpdatedDate": "2025-10-03T12:46:41.42Z",
                "commentType": "system",
            }
        ],
    }
    resources = [
        PullRequestThread.model_validate(spec),
        PullRequestThread.model_validate(spec | {"isDeleted": True}),
    ]
    assert not resources[0].deleted
    assert resources[1].deleted

    try:
        PullRequestThreadCollection.model_validate({"count": len(resources), "value": resources})
    except pydantic.ValidationError as exc:
        pytest.fail(reason=str(exc))

    with pytest.raises(pydantic.ValidationError):
        PullRequestThread.model_validate(spec | {"id": "foo"})


def test_git_pull_request_thread_body():
    """Test Git Pull Request Thread Body request instantiation."""
    resource = PullRequestThreadCreate(comments=[PullRequestThreadCommentCreate(content="Test")])
    assert resource.comments[0].content == "Test"

    with pytest.raises(pydantic.ValidationError):
        PullRequestThreadCreate(comments=1)
